@@ -1,54 +1,71 @@
 import logging
-logging.basicConfig(level=logging.DEBUG,format='%(asctime)s :: %(module)s :: %(levelname)s :: %(message)s')
-import clr
-logging.debug('Adding Reference Microsoft.AnalysisServices.AdomdClient')
-clr.AddReference('Microsoft.AnalysisServices.AdomdClient')
-logging.debug('Adding Reference Microsoft.AnalysisServices.Tabular')
-clr.AddReference('Microsoft.AnalysisServices.Tabular')
-logging.debug('Adding Reference Microsoft.AnalysisServices')
-clr.AddReference('Microsoft.AnalysisServices')
-logging.debug(f'Importing Microsoft.AnalysisServices.AdomdClient')
-from Microsoft.AnalysisServices.AdomdClient import AdomdCommand, AdomdConnection
+
 logging.debug(f'Importing Microsoft.AnalysisServices.Tabular')
 from Microsoft.AnalysisServices.Tabular import Server, Database, RefreshType, DataType, ConnectionDetails, ColumnType, MetadataPermission, Table, DataColumn, Partition, MPartitionSource, PartitionSourceType
+logging.debug(f'Importing Microsoft.AnalysisServices.AdomdClient')
+from Microsoft.AnalysisServices.AdomdClient import (AdomdCommand, AdomdConnection)
 logging.debug(f'Importing Microsoft.AnalysisServices')
 from Microsoft.AnalysisServices import UpdateOptions
 
 logging.debug('Importing Other Packages...')
-from typing import List
+from typing import List, Union
+from collections.abc import Iterable
 import requests as r
 import pandas as pd
 import json
 import os
 import subprocess
-from logic_utils import pd_dataframe_to_m_expression
+import atexit
+from logic_utils import pd_dataframe_to_m_expression, pandas_datatype_to_tabular_datatype
 
 class Tabular:
-	'''
-	Tabular Class. Input Connection String then you are off to the races...
-	This will be your best friend: https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular?view=analysisservices-dotnet
-	'''
-	def __init__(self,CONNECTION_STR):
+	'''Tabular Class to perform operations:[Microsoft.AnalysisServices.Tabular](https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular?view=analysisservices-dotnet)
+
+	Args:
+		CONNECTION_STR (str): [Connection String](https://docs.microsoft.com/en-us/analysis-services/instances/connection-string-properties-analysis-services?view=asallproducts-allversions)
+	'''	
+	def __init__(self,CONNECTION_STR:str):
 		logging.debug(f'Initializing Tabular Class')
-		self.Server = Server()
+		self.Server = Server() #[Server](https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.server?view=analysisservices-dotnet)
 		self.Server.Connect(CONNECTION_STR)
 		logging.debug(f'Connected to Server - {self.Server.Name}')
 		self.Catalog = self.Server.ConnectionInfo.Catalog
-		self.Database = self.Server.Databases.Find(self.Catalog)
+		logging.debug(f'Received Catalog - {self.Catalog}')
+		try:
+			self.Database = [database for database in self.Server.Databases.GetEnumerator() if database.Name == self.Catalog][0]
+		except:
+			logging.error(f'Unable to find Database... {self.Catalog}')
 		logging.debug(f'Connected to Database - {self.Database.Name}')
+		self.CompatibilityLevel: int = self.Database.CompatibilityLevel
+		self.CompatibilityMode: int = self.Database.CompatibilityMode.value__
 		self.Model = self.Database.Model
+		logging.debug(f'Connected to Model - {self.Model.Name}')
 		self.DaxConnection = AdomdConnection()
 		self.DaxConnection.ConnectionString = f"{self.Server.ConnectionString}Password='{self.Server.ConnectionInfo.Password}'"
+		self.Reload_Model_Info()
+		logging.debug(f'Class Initialization Completed')
+		logging.debug(f'Registering Disconnect on Termination...')
+		atexit.register(self.Disconnect)
+		
+		pass
+	def __repr__(self) -> str:
+		return f'{self.Server.Name}::{self.Database.Name}::{self.Model.Name}\n{self.Database.EstimatedSize} Estimated Size\n{len(self.Tables)} Tables\n{len(self.Columns)} Columns\n{len(self.Partitions)} Partitions\n{len(self.Measures)} Measures'
+	def Reload_Model_Info(self) -> bool:
+		'''Runs on __init__ iterates through details, can be called after any model changes. Called in SaveChanges()
+
+		Returns:
+			bool: True if successful
+		'''
 		self.Tables = [table for table in self.Model.Tables.GetEnumerator()]
 		self.Columns = [column for table in self.Tables for column in table.Columns.GetEnumerator()]
 		self.Partitions = [partition for table in self.Tables for partition in table.Partitions.GetEnumerator()]
 		self.Measures = [measure for table in self.Tables for measure in table.Measures.GetEnumerator()]
-		logging.debug(f'Class Initialization Completed')
-		pass
+		return True
 	def Disconnect(self) -> bool:
-		'''
-		Self explanatory. Runs self.Server.Disconnect()
-		For logging purposes will check if disconnect successful.
+		'''Disconnects from Model
+
+		Returns:
+			bool: True if successful
 		'''
 		logging.debug(f'Disconnecting from - {self.Server.Name}')
 		self.Server.Disconnect()
@@ -59,26 +76,59 @@ class Tabular:
 		else:
 			logging.debug(f'Disconnect Successful')
 			return True	
-	def Refresh(self, iterable_items: List, RefreshType=RefreshType.Full) -> None:
+	def Refresh(self, Object:Union[str,Table,Partition,Iterable], RefreshType=RefreshType.Full) -> None:
+		'''Input Object(s) to be refreshed in the tabular model. Combine with .SaveChanges() to actually run the refresh on the model.
+
+		Args:
+			Object (Union[str,Table,Partition,Iterable]): Can be str(table name only), Table object, Partition object, or an iterable combination of the three.
+			RefreshType (_type_, optional): [RefreshType](https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.refreshtype?view=analysisservices-dotnet). Defaults to RefreshType.Full.
 		'''
-		Input iterable Collections for the function to run through.
-		It will add the collection items into a Refresh Request.
-		To execute refresh run through Update()
-		'''
-		for collection in iterable_items:
-			logging.debug(f'Adding {collection.Name} to Refresh Request')
-			collection.RequestRefresh(RefreshType)
-	def Update(self, UpdateOptions=UpdateOptions.ExpandFull) -> None:
-		'''
-		Takes currently changed options in model then updates.
+		logging.debug(f'Beginning RequestRefresh cadence...')
+		def refresh(object):
+			if isinstance(object,str):
+				logging.info(f'Requesting refresh for {object}')
+				table = [table for table in self.Tables if table.Name == object][0]
+				table.RequestRefresh(RefreshType)
+			else:
+				logging.info(f'Requesting refresh for {object.Name}')
+				object.RequestRefresh(RefreshType)
+		if isinstance(Object,Iterable):
+			[refresh(object) for object in Object]
+		else:
+			refresh(Object)
+			Object.RequestRefresh(RefreshType)
+	def Update(self, UpdateOptions:UpdateOptions =UpdateOptions.ExpandFull) -> None:
+		'''[Update Model](https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.majorobject.update?view=analysisservices-dotnet#microsoft-analysisservices-majorobject-update(microsoft-analysisservices-updateoptions))
+
+		Args:
+			UpdateOptions (UpdateOptions, optional): See above MS Doc link. Defaults to UpdateOptions.ExpandFull.
+
+		Returns:
+			None: Placeholder to eventually change.
 		'''
 		logging.debug('Running Update Request')
 		return self.Database.Update(UpdateOptions)
-	def Backup_Table(self,table_str:str = 'Table Name'):
-		'''
-		This will copy and refresh a table appending _backup to the end of the names for every object
-		It will bring over every relationship, role, measure, column, etc...
+	def SaveChanges(self) -> bool:
+		'''TODO need to clean this up and add more flexibility.  
+		Just a simple wrapper to call self.Model.SaveChanges()
+
+		Returns:
+			bool:
 		'''
+		self.Reload_Model_Info()
+		self.Model.SaveChanges()
+		return True
+	def Backup_Table(self,table_str:str) -> bool:
+		'''USE WITH CAUTION, EXPERIMENTAL. Backs up table in memory, brings with it measures, columns, hierarchies, relationships, roles, etc.  
+		It will add suffix '_backup' to all objects.  
+		Refresh is performed from source during backup.
+
+		Args:
+			table_str (str, optional): Name of Table.
+
+		Returns:
+			bool: Returns True if Successful, else will return error.
+		'''		
 		logging.info('Backup Beginning...')
 		logging.debug(f'Cloning {table_str}')
 		table = self.Model.Tables.Find(table_str).Clone()
@@ -133,12 +183,25 @@ class Tabular:
 		logging.info(f'Refreshing Clone... {table.Name}')
 		self.Refresh([table])
 		logging.info(f'Updating Model {self.Model.Name}')
-		self.Update()
+		self.SaveChanges()
 		return True
-	def Revert_Table(self, table_str:str = 'Table Name'):
-		'''
-		This is used in conjunction with Backup_Table().
+	def Revert_Table(self, table_str:str) -> bool:
+		'''USE WITH CAUTION, EXPERIMENTAL. This is used in conjunction with Backup_Table().
 		It will take the 'TableName_backup' and replace with the original.
+		Example scenario -> 
+		1. model.Backup_Table('TableName')
+		2. perform any proposed changes in original 'TableName'
+		3. validate changes in 'TableName'
+		4. if unsuccessful run model.Revert_Table('TableName')
+
+		Args:
+			table_str (str): Name of table.
+
+		Returns:
+			bool: Returns True if Successful, else will return error.
+		'''		
+		'''
+
 		'''
 		logging.info(f'Beginning Revert for {table_str}')
 		logging.debug(f'Finding original {table_str}')
@@ -188,11 +251,16 @@ class Tabular:
 		dename(backup_relationships)
 		logging.info(f'Name changes for Backup Table...')
 		backup.RequestRename(backup.Name.removesuffix('_backup'))
-		self.Update()
+		self.SaveChanges()
 		return True
 	def Query(self,Query_Str:str) -> pd.DataFrame:
-		'''
-		Executes Query on Model and Returns Results in Pandas DataFrame
+		'''	Executes Query on Model and Returns Results in Pandas DataFrame
+
+		Args:
+			Query_Str (str): Dax Query. Note, needs full syntax (ex: EVALUATE). See https://docs.microsoft.com/en-us/dax/dax-queries 
+
+		Returns:
+			pd.DataFrame: Returns dataframe with results
 		'''
 		logging.info(f'Query Called...')
 		try:
@@ -215,10 +283,15 @@ class Tabular:
 		logging.debug(f'Converting to Pandas DataFrame...')
 		df = pd.DataFrame(Results,columns=[value for _,value in Column_Headers])
 		return df
-	def Query_Every_Column(self,query_function:str='COUNTROWS(VALUES(_))') -> pd.DataFrame():
-		'''
-		This will dynamically create a query to pull all columns from the model and run the query function.
-		It will replace the _ with the column to run.
+	def Query_Every_Column(self,query_function:str='COUNTROWS(VALUES(_))') -> pd.DataFrame:
+		'''This will dynamically create a query to pull all columns from the model and run the query function.
+		<br/>It will replace the _ with the column to run.
+
+		Args:
+			query_function (str, optional): Dax query is dynamically building a query with the UNION & ROW DAX Functions.
+
+		Returns:
+			pd.DataFrame: Returns dataframe with results.
 		'''
 		logging.info(f'Beginning execution of querying every column...')
 		logging.debug(f'Function to be run: {query_function}')
@@ -232,10 +305,15 @@ class Tabular:
 				query_str += f"ROW(\"Table\",\"{table_name}\",\"Column\",\"{column_name}\",\"{query_function}\",{query_function.replace('_',dax_identifier)}),\n"
 		query_str = f'{query_str[:-2]})'
 		return self.Query(query_str)
-	def Query_Every_Table(self,query_function:str='COUNTROWS(_)') -> pd.DataFrame():
-		'''
-		This will dynamically create a query to pull all tables from the model and run the query function.
+	def Query_Every_Table(self,query_function:str='COUNTROWS(_)') -> pd.DataFrame:
+		'''This will dynamically create a query to pull all tables from the model and run the query function.
 		It will replace the _ with the table to run.
+
+		Args:
+			query_function (str, optional): Dax query is dynamically building a query with the UNION & ROW DAX Functions. Defaults to 'COUNTROWS(_)'.
+
+		Returns:
+			pd.DataFrame: Returns dataframe with results
 		'''
 		logging.info(f'Beginning execution of querying every table...')
 		logging.debug(f'Function to be run: {query_function}')
@@ -247,7 +325,18 @@ class Tabular:
 			query_str += f"ROW(\"Table\",\"{table_name}\",\"{query_function}\",{query_function.replace('_',dax_table_identifier)}),\n"
 		query_str = f'{query_str[:-2]})'
 		return self.Query(query_str)
-	def Analyze_BPA(self,Tabular_Editor_Exe,Best_Practice_Analyzer):
+	def Analyze_BPA(self,Tabular_Editor_Exe:str,Best_Practice_Analyzer:str) -> List[str]:
+		'''Takes your Tabular Model and performs TE2s BPA. Runs through Command line.
+		https://docs.tabulareditor.com/te2/Best-Practice-Analyzer.html
+		https://docs.tabulareditor.com/te2/Command-line-Options.html
+
+		Args:
+			Tabular_Editor_Exe (str): TE2 Exe File path. Feel free to use class TE2().EXE_Path or provide your own.
+			Best_Practice_Analyzer (str): BPA json file path. Feel free to use class BPA().Location or provide your own. Defualts to 	https://raw.githubusercontent.com/microsoft/Analysis-Services/master/BestPracticeRules/BPARules.json
+
+		Returns:
+			List[str]: Assuming no failure, will return list of BPA violations. Else will return error from command line.
+		'''		
 		#Working TE2 Script in Python os.system(f"start /wait {te2.EXE_Path} \"Provider=MSOLAP;{model.DaxConnection.ConnectionString}\" FINANCE -B \"{os.getcwd()}\\Model.bim\" -A {l.BPA_LOCAL_FILE_PATH} -V/?")
 		#start /wait 
 		logging.debug(f'Beginning request to talk with TE2 & Find BPA...')
@@ -260,22 +349,30 @@ class Tabular:
 			return error
 		else:
 			return [output for output in raw_output.split('\n') if 'violates rule' in output]
-	def Create_Table(self,df:pd.DataFrame = pd.DataFrame(data={'col1': [1.0, 2.0], 'col2': [3, 4]}),table_name:str = 'Testing') -> bool:
-		'''
-		Should create table as m-partition from a pandas df
-		'''
+	def Create_Table(self,df:pd.DataFrame, table_name:str) -> bool:
+		'''Creates tables from pd.DataFrame as an M-Partition. 
+		So will convert the dataframe to M-Partition logic via the M query table constructor.
+		Runs refresh and will update model.
+
+		Args:
+			df (pd.DataFrame): DataFrame to add to model
+			table_name (str): _description_
+
+		Returns:
+			bool: True if successful
+		'''	
 		logging.debug(f'Beginning to create table for {table_name}...')
 		new_table = Table()
 		new_table.RequestRename(table_name)
 		logging.debug(f'Sorting through columns...')
 		df_column_names = df.columns
-		#\u007b = { and \u007d = }
+		dtype_conversion = pandas_datatype_to_tabular_datatype(df)
 		for df_column_name in df_column_names:
 			logging.debug(f'Adding {df_column_name} to Table...')
 			column = DataColumn()
 			column.RequestRename(df_column_name)
 			column.set_SourceColumn(df_column_name)
-			column.set_DataType(DataType.String)
+			column.set_DataType(dtype_conversion[df_column_name])
 			new_table.Columns.Add(column)
 		logging.debug(f'Expression String Created...')
 		logging.debug(f'Creating MPartition...')
@@ -288,17 +385,16 @@ class Tabular:
 		logging.debug(f'Adding table: {new_table.Name} to {self.Server.Name}::{self.Database.Name}::{self.Model.Name}')
 		self.Model.Tables.Add(new_table)
 		self.Refresh([new_table])
-		self.Update()
+		self.SaveChanges()
 		return True
 
 
 
 class BPA:
-	'''
-	Best Practice Analyzer Class 
-		Can provide Url, Json File Path, or Python List.
-		If nothing is provided it will default to Microsofts Analysis Services report with BPA Rules.
-		https://raw.githubusercontent.com/microsoft/Analysis-Services/master/BestPracticeRules/BPARules.json
+	'''_summary_
+	'''	
+	'''Best Practice Analyzer Class. Can provide Url, Json File Path, or Python List. If nothing is provided it will default to Microsofts Analysis Services report with BPA Rules. 
+	[Default BPA](https://raw.githubusercontent.com/microsoft/Analysis-Services/master/BestPracticeRules/BPARules.json)
 	'''
 	def __init__(self,rules_location:str='https://raw.githubusercontent.com/microsoft/Analysis-Services/master/BestPracticeRules/BPARules.json') -> None:
 		'''
@@ -317,13 +413,11 @@ class BPA:
 				self.Rules = json.load(json_file)
 			logging.debug(f'Rules from file path collected...')
 		pass
-
+#Todo... subclass with a namedtuple
 class TE2:
-	'''
-	TE2 Class, to use any built TabularEditor Command Line Scripts
-	https://docs.tabulareditor.com/te2/Command-line-Options.html
-	#https://github.com/TabularEditor/TabularEditor/releases/download/2.16.7/TabularEditor.Portable.zip
-	#https://cdn.tabulareditor.com/files/TabularEditor.2.16.7.zip
+	'''TE2 Class, to use any built TabularEditor Command Line Scripts  
+	[TE2 Command Line Example](https://docs.tabulareditor.com/te2/Command-line-Options.html)  
+	[TE2 Download](https://github.com/TabularEditor/TabularEditor/releases/download/2.16.7/TabularEditor.Portable.zip)
 	'''
 	def __init__(self,TE_Location='https://github.com/TabularEditor/TabularEditor/releases/download/2.16.7/TabularEditor.Portable.zip') -> None:
 		logging.debug(f'Checking for TE2 in {os.getcwd()}')
